/*
-----------------------------------------------------------------------------
This source file is part of the GameTools Project
http://www.gametools.org

Author: Martin Szydlowski
-----------------------------------------------------------------------------
*/

#ifndef _KdTreeSceneManager_H__
#define _KdTreeSceneManager_H__

#include <OgreKdTreeHierarchyInterface.h>
#include <OgreSceneManager.h>
#include <VisibilityManager.h>
#include "OgreKdTree.h"

#define KDTREE_MAX_DEPTH 20

namespace Ogre
{
	class KdTreeSceneNode;
	class WireBoundingBox;
	//typedef std::list < WireBoundingBox * > WBBList;

	class KdTreeSceneManager : public SceneManager
	{
	public:
		KdTreeSceneManager(const String& name, GtpVisibility::VisibilityManager *vm);
		~KdTreeSceneManager(void);

		virtual const String& getTypeName(void) const;

		/** Override from SceneManager to create SceneNodes as instance of KdTreeSceneNode
		*/
		virtual SceneNode* createSceneNode(void);
		/** Override from SceneManager to create SceneNodes as instance of KdTreeSceneNode
		*/
		virtual SceneNode* createSceneNode(const String& name);

		/** Override from SceneManager so that sub entities can be assigned an id for item buffer.
		*/
		Entity* createEntity(const String& entityName, const String& meshName);

		/** Update the KdTree with the node (more or less disabled now)
		*/
		virtual void _updateNode(KdTreeSceneNode *node); // TODO: 

		/** Override from scene manager to st up culling manager
		*/
		virtual void _updateSceneGraph(Camera* cam);
		/** Override from Scenemanager, employ kd-tree based culling
			or CHC
		*/
		virtual void _findVisibleObjects(Camera *cam, bool onlyShadowCasters);
		//virtual void _renderVisibleObjects();

		/** Render a list of scenenodes
		*/
		virtual void _renderNodes(const KdRenderableList& nodelist, Camera * cam,
			bool onlyShadowCasters, int leavePassesInQueue);

		virtual void setShowBoxes(bool showboxes);
		virtual bool getShowBoxes(void) const;

		virtual bool setOption(const String& strKey, const void* pValue);
		virtual bool getOption(const String& strKey, void* pDestValue);

		bool getOptionValues(const String & key, StringVector &refValueList);
		bool getOptionKeys(StringVector &refKeys);

		/** Sets the visibility manager.
		@param visManager the visibility manager
		*/
		void setVisibilityManager(GtpVisibility::VisibilityManager *visManager);
		/** See set.
		*/
		GtpVisibility::VisibilityManager *getVisibilityManager();

		/** Returns pointer to visibility manager.
		*/
		GtpVisibility::VisibilityManager *GetVisibilityManager();

		/** Returns hierarchy interface.
		*/
		KdTreeHierarchyInterface *GetHierarchyInterface();


		///** Render a queue group. 
		//Override so we can handle delayed rendering of transparent objects
		//*/
		//void renderBasicQueueGroupObjects(RenderQueueGroup* pGroup, 
		//	QueuedRenderableCollection::OrganisationMode om);

		///** Writes out stats into the Ogre log file.
		//*/
		//void WriteLog();

		///** Override pass so we can do the z-fail pass. 
		//*/
		//const Pass* _setPass(Pass* pass);

		///** Override from SceneManager so we can skip all but first pass for depth pass. 
		//*/
		//bool validatePassForRendering(Pass* pass);

		virtual void clearScene()
		{
			// DEBUG must happen before scene is cleared
			// might be a solution to detach problem
			//if (mKdTree)
			//	mKdTree->dump();
			
			OGRE_DELETE(mKdTree);

			SceneManager::clearScene();
		}

	protected:
		GtpVisibility::VisibilityManager *mVisibilityManager;

		KdTreeHierarchyInterface *mHierarchyInterface;

		/// consecutive number for sub-entities
		int mCurrentEntityId;

		/// flag for passes which should not be deleted from queue during first traversal
		int mLeavePassesInQueue;
		
		int mMaxDepth;

		KdTree *mKdTree;

		bool mShowBoxes;

		KdTree::RenderMethod mRenderMethod;

		KdTree::BuildMethod mBuildMethod;
#ifdef KDTREE_DEBUG
		int mHighlighLevel;
		bool mShowAllBoxes;
#endif
	};



	/// Factory for KdTreeSceneManager
	class KdTreeSceneManagerFactory : public SceneManagerFactory
	{
	protected:
		void initMetaData(void) const;
		GtpVisibility::VisibilityManager *visManager;
	public:
		KdTreeSceneManagerFactory(GtpVisibility::VisibilityManager * vm)
		{
			visManager = vm;
		};
		~KdTreeSceneManagerFactory(void) {};
		/// Factory type name
		static const String FACTORY_TYPE_NAME;
		SceneManager* createInstance(const String& instanceName);
		void destroyInstance(SceneManager* instance);
	};

} // namespace Ogre

#endif // _KdTreeSceneManager_H__